-- Start Code --

-- Light/Fan Button Timer
-- Copyright 2022 Wirepath Home Systems, LLC. All rights reserved.

-- MAX values must match items found in properties in driver .xml
MAX_COLORS = 3
MAX_TIMERS = 9

-- FSC button IDs for LEDs
BUTTON = {HIGH = 0, MED_HIGH = 1, MEDIUM = 2, LOW = 3, OFF = 4}
BUTTONMAP = {}

-- Gen1 Top/Bottom buttons
BUTTONMAP[1] = "Top"
BUTTONMAP[3] = "Bottom"

-- Gen3 Top/Bottom buttons
BUTTONMAP[5002] = "Top"
BUTTONMAP[5007] = "Bottom"

-- FSC button maps
BUTTONMAP[5001] = "High"
BUTTONMAP[5101] = "Medium High"
BUTTONMAP[5201] = "Medium"
BUTTONMAP[5301] = "Low"
BUTTONMAP[5401] = "Off"

-- Reverse Buttom Map.
EVENTMAP = {}
EVENTMAP["Gen1 Top"]        = 1
EVENTMAP["Gen1 Bottom"]     = 3
EVENTMAP["Gen3 Top"]        = 5002
EVENTMAP["Gen3 Bottom"]     = 5007
EVENTMAP["FSC High"]        = 5001
EVENTMAP["FSC Medium High"] = 5101
EVENTMAP["FSC Medium"]      = 5201
EVENTMAP["FSC Low"]         = 5301
EVENTMAP["FSC Off"]         = 5401

-- Button ID Mapping
BUTTONIDMAP = {}
BUTTONIDMAP[5001] = 0
BUTTONIDMAP[5101] = 1
BUTTONIDMAP[5201] = 2
BUTTONIDMAP[5301] = 3
BUTTONIDMAP[5401] = 4

-- Mapping of Off ramp times to seconds.
OFFMODES = {}
OFFMODES["Off"] = 0
OFFMODES["Ramp 1 Second"]   =  1
OFFMODES["Ramp 5 Seconds"]  =  5
OFFMODES["Ramp 10 Seconds"] = 10
OFFMODES["Ramp 30 Seconds"] = 30
OFFMODES["Ramp 1 Minute"]   =  1 * 60
OFFMODES["Ramp 5 Minutes"]  =  5 * 60
OFFMODES["Ramp 10 Minutes"] = 10 * 60
OFFMODES["Ramp 30 Minutes"] = 30 * 60

-- Hidden/Visible property enum
PROP_VISIBLE, PROP_HIDDEN = 0, 1


function OnPropertyChanged(strProperty)
  local prop = Properties[strProperty] or ""

  if (strProperty == "Debug Mode") then
    gDbgTimer = C4:KillTimer(gDbgTimer or 0)
    gDbgPrint, gDbgLog = (prop:find("Print") ~= nil), (prop:find("Log") ~= nil)
    if (prop == "Off") then return end

    gDbgTimer = C4:AddTimer(8*60, "MINUTES")
    dbg("Enabled Debug Timer for 8 Hours")
    return
  end

  -- Scut work of hiding / showing various properties...
  -- Post Driver Late Init, each 'Timer X' Property is triggered, to re-show/hide.
  if (strProperty:find("^Timer %d$") ~= nil) then
    local _, _, idTimer = strProperty:find("^Timer (%d)$")
    if (prop == "Off") then
      C4:SetPropertyAttribs("Timer " .. idTimer .. " Devices", PROP_HIDDEN)
      C4:SetPropertyAttribs("Timer " .. idTimer .. " Times", PROP_HIDDEN)
      C4:SetPropertyAttribs("Timer " .. idTimer .. " Units", PROP_HIDDEN)
      C4:SetPropertyAttribs("Timer " .. idTimer .. " Off Mode", PROP_HIDDEN)
      for i = 1, MAX_COLORS do
        C4:SetPropertyAttribs("Timer " .. idTimer .. " Color " .. i, PROP_HIDDEN)
      end
      C4:SetPropertyAttribs("Timer " .. idTimer .. " On Color", PROP_HIDDEN)
    else
      C4:SetPropertyAttribs("Timer " .. idTimer .. " Devices", PROP_VISIBLE)
      C4:SetPropertyAttribs("Timer " .. idTimer .. " Times", PROP_VISIBLE)
      C4:SetPropertyAttribs("Timer " .. idTimer .. " Units", PROP_VISIBLE)
      C4:SetPropertyAttribs("Timer " .. idTimer .. " Off Mode", PROP_VISIBLE)
      C4:SetPropertyAttribs("Timer " .. idTimer .. " On Color", PROP_VISIBLE)

      for i = 1, MAX_COLORS do
        if (prop:find(tostring(i))) then
          C4:SetPropertyAttribs("Timer " .. idTimer .. " Color " .. i, PROP_VISIBLE)
        else
          C4:SetPropertyAttribs("Timer " .. idTimer .. " Color " .. i, PROP_HIDDEN)
        end
      end
    end
  end

  -- Un-register and Re-register all Device id/events on *any* 'Timer X Devices' property changed.
  -- Safest way to ensure they're only registered to one timer...
  -- Post Driver Late Init, 'Timer 1 Devices' Property is triggered, to re-reigster device events.

  -- Supported Devices...
  local SUPPORTED_DEVICES = {
    ["light_ip_control4_ldz-5s1-w.c4i"] = "GEN1 SWITCH", -- Gen1 Switch
    ["switch_gen3.c4i"]                 = "GEN3 SWITCH", -- Gen3 Switch
    ["light_ip_control4_ldz-101-w.c4i"] = "GEN1 DIMMER", -- Gen1 Dimmer
    ["adaptive_phase_dimmer.c4i"]       = "GEN3 DIMMER", -- Gen3 Dimmer
    ["forward_phase_dimmer.c4i"]        = "GEN3 DIMMER", -- Gen3 Dimmer
    ["fan_speed_controller.c4i"]        = "FSC",         -- FSC
  }

  if (strProperty:find("^Timer %d Devices$") ~= nil) then

    C4:UnregisterAllDeviceEvents()
    gDev = {}

    for propname,_ in pairs(Properties) do
      if (propname:find("^Timer %d Devices$") ~= nil) then

        prop = trim(Properties[propname]) .. ","
        for devid in prop:gfind("(%d+),") do
          local idDevice = tonumber(devid) or 0

          if (idDevice ~= 0) then
            local dev_name = C4:GetDeviceDisplayName(idDevice) or "UNKNOWN"

            if (gDev[idDevice] ~= nil) then  -- Device can only be registered to a single timer...
              if (gDev[idDevice].idTimer) then
                dbg("Device id: " .. idDevice .. " (" .. dev_name .. ") is already registered to Timer " .. gDev[idDevice].idTimer)
              end
            else
              local _, _, idTimer = propname:find("^Timer (%d) Devices$")

              gDev[idDevice] = {}
              gDev[idDevice].idTimer = idTimer

              local proto_device = ((C4:GetDevices({}, idDevice) or {})[idDevice] or {}).protocol
              local _, t = next(proto_device)
              local device_type = "UNKNOWN"
              local protocol_dev = "UNKNOWN"
              if (type(t) == "table") then
                protocol_dev = t.driverFileName
                device_type = SUPPORTED_DEVICES[protocol_dev] or "UNKNOWN"
              end

              if (device_type == "UNKNOWN") then
                dbg("Timer " .. idTimer .. " Not Registering unsupported device type (" .. protocol_dev .. "): " .. "[" .. idDevice .. "]: " ..
                    C4:GetDeviceDisplayName(idDevice))
              else

                if (device_type == "GEN1 DIMMER") then
                  C4:RegisterDeviceEvent(idDevice, EVENTMAP["Gen1 Top"])
                  C4:RegisterDeviceEvent(idDevice, EVENTMAP["Gen1 Bottom"])
                  gDev[idDevice].devType = "Dimmer"
                  gDev[idDevice].devGen = "Gen1"

                elseif (device_type == "GEN3 DIMMER") then
                  C4:RegisterDeviceEvent(idDevice, EVENTMAP["Gen3 Top"])
                  C4:RegisterDeviceEvent(idDevice, EVENTMAP["Gen3 Bottom"])
                  gDev[idDevice].devType = "Dimmer"
                  gDev[idDevice].devGen = "Gen3"

                elseif (device_type == "GEN1 SWITCH") then
                  C4:RegisterDeviceEvent(idDevice, EVENTMAP["Gen1 Top"])
                  C4:RegisterDeviceEvent(idDevice, EVENTMAP["Gen1 Bottom"])
                  gDev[idDevice].devType = "Switch"
                  gDev[idDevice].devGen = "Gen1"

                elseif (device_type == "GEN3 SWITCH") then
                  C4:RegisterDeviceEvent(idDevice, EVENTMAP["Gen3 Top"])
                  C4:RegisterDeviceEvent(idDevice, EVENTMAP["Gen3 Bottom"])
                  gDev[idDevice].devType = "Switch"
                  gDev[idDevice].devGen = "Gen3"

                elseif (device_type == "FSC") then
                  gDev[idDevice+1] = {}
                  gDev[idDevice+1].idTimer = idTimer
                  gDev[idDevice+1].devType = "FSC"
                  gDev[idDevice+1].devGen = "Gen3"

                  -- FSC Keypad 5001 == High, 5101 == Med-High, 5201 == Medium, 5301 == Low, 5401 == Off.
                  -- Events are registered on the Keypad Proxy, not on the FSC proxy or protocol driver.
                  for k,_ in pairs(BUTTONIDMAP) do
                    C4:RegisterDeviceEvent(idDevice+1, k)
                  end
                end
                dbg("Timer " .. idTimer .. " Registering " .. device_type .. " id: " .. idDevice .. " (" .. dev_name .. ")")
              end
            end
          end
        end
      end
    end
  end
end


function dbg(strDebugText)
  if (gDbgPrint) then print(strDebugText) end
  if (gDbgLog) then C4:DebugLog("\r\n" .. strDebugText) end
end


function OnDriverDestroyed()
  gDbgTimer = C4:KillTimer(gDbgTimer or 0)
  C4:UnregisterAllDeviceEvents()
end


function OnDriverLateInit()
  if (LuaC4Object.AllowExecute) then C4:AllowExecute(true) end

  OneShotTimer.Add(5, "SECONDS",
    function()
      -- Force Show/Hide for all properties...
      for i = 1, MAX_TIMERS do OnPropertyChanged("Timer " .. i) end

	    -- Cause re-registration of devices' events by triggering a single 'Devices' property.
      OnPropertyChanged("Timer 1 Devices")
    end
    , "ODLI"
  )
end


function trim(s)
  return s:gsub("^%s*(.-)%s*$", "%1") -- Trim Whitespace...
end


function ExecuteCommand(strCommand, _) -- tParams
  dbg("ExecuteCommand: " .. strCommand)
end


function MySendToDevice(idDevice, strCommand, tParams)
  tParams = tParams or {}
  idDevice = idDevice or 0
  local DEBUG_MAXLEN = 40

  local out = {}
  table.insert(out, "----> C4:SendToDevice(" .. idDevice .. ", ")
  table.insert(out, strCommand .. ", {")
  for k,v in pairs(tParams) do
    if (type(v) == "string") then
      table.insert(out, k .. ' = "' .. (tostring(v)):sub(1, DEBUG_MAXLEN) .. '",')
    else
      table.insert(out, k .. " = " .. (tostring(v)):sub(1, DEBUG_MAXLEN) .. ",")
    end
  end
  table.insert(out, "}")
  dbg(table.concat(out, " "):gsub(", }", " }") .. ")")

  C4:SendToDevice(idDevice, strCommand, tParams)
end


function GetNext(timer, curmode, modes, times)
  curmode = (curmode or 0) + 1
  if ((modes[curmode] ~= nil) and (modes[curmode] ~= "On")) then
    return curmode, times[curmode], Properties["Timer " .. timer .. " Color " .. curmode]
  else
    if (modes[curmode] == nil) then
      return curmode-1, 0, ""
    else
      return curmode, 0, ""
    end
  end
end


function tohexcolor(color)
  local colors, out = ParseCSV(color), {}
  for _,v in pairs(colors) do
    v = tonumber(v) or 0
    table.insert(out, string.format("%02X", v))
  end
  return table.concat(out, "")
end


function SetButtonLEDColor(idDevice, color, button)
  button = button or 0
  color = tohexcolor(color)
  local dev = gDev[idDevice]

  if (dev.devType == "FSC") then
    MySendToDevice(idDevice, "KEYPAD_BUTTON_COLOR", {ON_COLOR = color, BUTTON_ID = button})
  else
    if (dev.devGen == "Gen1") then
      MySendToDevice(idDevice, "SET_TOP_LED_ON", {RGB = color})
    end
    if (dev.devGen == "Gen3") then
      MySendToDevice(idDevice, "SET_BUTTON_COLOR", {ON_COLOR = color, BUTTON_ID = button})
    end
  end
end


function SetAllLEDs_FSC(idDevice, color)
  for i = BUTTON.HIGH, BUTTON.OFF do
    SetButtonLEDColor(idDevice, color, i)
  end
end


function ParseCSV(csv)
  local out = {}
  csv = csv:gsub("%s", "") .. ","
  csv:gsub("(.-),", function(t) table.insert(out, t) end)
  return out
end


function AdvanceFSCMode(idDevice, idTimer, button)
  local tm, color
  local mode = Properties["Timer " .. idTimer]
  local time = Properties["Timer " .. idTimer .. " Times"] or 0
  local units = Properties["Timer " .. idTimer .. " Units"]
  local modes, times = ParseCSV(mode), ParseCSV(time)
  local on_color = Properties["Timer " .. idTimer .. " On Color"]
  local dev = gDev[idDevice]

  -- Only go to 'next' on same button press, or first press (prev was Off)...
  if (dev.curbutton == button) or (dev.curbutton == BUTTON.OFF) then
    local lastmode = dev.curmode or 'nil'
    dev.curmode, tm, color = GetNext(idTimer, dev.curmode, modes, times)
    dev.curcolor = color
    if (tostring(tm) == "") then return end
    if (tm ~= 0) then
      dbg("Timer " .. idTimer .. " [" .. idDevice .. "] Mode: " .. dev.curmode .. " Time: " .. tm .. " " .. units ..
          " Color: " .. color .. " (" .. tohexcolor(color) .. ")")

      -- Set 'Turn Off' timer for this device...
      OneShotTimer.Add(tm, units:upper(),
        function()
	        SetAllLEDs_FSC(idDevice, on_color)
	        MySendToDevice(idDevice, "OFF", {})
	        dev.curmode = nil
	        dev.curbutton = BUTTON.OFF
        end
        , idDevice
      )
    else
      if (lastmode ~= dev.curmode) then    -- We're in 'ON' Mode
        dbg("Timer " .. idTimer .. " [" .. idDevice .. "] Mode: " .. dev.curmode .. " Time: ON")
        SetButtonLEDColor(idDevice, on_color, button) -- Reset LED for this device...
        OneShotTimer.Kill(idDevice) -- Kill Timer for this device...
        dev.curbutton = button
        dev.curcolor = on_color
      end
      return
    end
  end
  SetButtonLEDColor(idDevice, dev.curcolor, button)
  dev.curbutton = button
end


function OnDeviceEvent(idDevice, idEvent)
  dbg("<---- Event[" .. idDevice .. "]: " .. BUTTONMAP[idEvent] .. " Button Pressed")
  local dev = gDev[idDevice]
  local idTimer = dev.idTimer
  local mode = Properties["Timer " .. idTimer]
  local time = Properties["Timer " .. idTimer .. " Times"]
  local units = Properties["Timer " .. idTimer .. " Units"]
  local offmode = Properties["Timer " .. idTimer .. " Off Mode"]
        offmode = OFFMODES[offmode] or 0
  local on_color = Properties["Timer " .. idTimer .. " On Color"]
  local modes, times = ParseCSV(mode), ParseCSV(time)
  local tm, color

  if (mode == "Off") then dbg("Timer " .. idTimer .. " [" .. idDevice .. "] Mode: (Timer Off)") return end -- This timer is 'turned off'.

  if (idEvent == EVENTMAP["Gen1 Top"]) or (idEvent == EVENTMAP["Gen3 Top"]) then -- Top Button pressed

    local lastmode = dev.curmode or 'nil'
    dev.curmode, tm, color = GetNext(idTimer, dev.curmode, modes, times)
    tm = tm or 0

    if (tostring(tm) == "") then return end
    if (tm ~= 0) then
      dbg("Timer " .. idTimer .. " [" .. idDevice .. "] Mode: " .. dev.curmode .. " Time: " .. tm .. " " .. units ..
          " Color: " .. color .. " (" .. tohexcolor(color) .. ")")

      SetButtonLEDColor(idDevice, color) -- Set LED for this device...
      -- Set 'Turn Off' timer and LED reset for this device...

      OneShotTimer.Add(tm, units:upper(),
        function()
	        SetButtonLEDColor(idDevice, on_color) -- Reset LED for this device...
          dev.curmode = nil

          local devtype = dev.devType or "Switch"
          if (devtype == "Switch") or (offmode == 0) then
            MySendToDevice(idDevice, "OFF", {})
          else
            MySendToDevice(idDevice, "RAMP_TO_LEVEL", {TIME = offmode * 1000, LEVEL = 0})
          end
        end
        , idDevice
      )

    else
      if (lastmode ~= dev.curmode) then -- We're in 'ON' Mode
        dbg("Timer " .. idTimer .. " [" .. idDevice .. "] Mode: " .. dev.curmode .. " Time: ON")
        SetButtonLEDColor(idDevice, on_color) -- Reset LED for this device...
        OneShotTimer.Kill(idDevice) -- Kill Timer for this device...
      end
    end
    return
  end

  if (idEvent == EVENTMAP["Gen1 Bottom"]) or (idEvent == EVENTMAP["Gen3 Bottom"]) then -- Bottom Button pressed
    SetButtonLEDColor(idDevice, on_color) -- Reset LED for this device...
    OneShotTimer.Kill(idDevice) -- Kill Timer for this device...
    dev.curmode = nil
    return
  end

  dev.curbutton = dev.curbutton or BUTTON.OFF

  if (idEvent >= EVENTMAP["FSC High"]) and (idEvent <= EVENTMAP["FSC Low"]) then
    AdvanceFSCMode(idDevice, idTimer, BUTTONIDMAP[idEvent])
    return
  end

  if (idEvent == EVENTMAP["FSC Off"]) then
    SetAllLEDs_FSC(idDevice, on_color)
    OneShotTimer.Kill(idDevice) -- Kill Timer for this device...
    dev.curmode = nil
    dev.curbutton = nil
    return
  end
end


----------------------- INCLUDES ---------------------------
-- TODO: Make this a true module, by returning it inside a local table...
-- Now allows for 'named' one-shot timers.  If the name is re-used, reset timer, don't make a new one...

---------------------------------------------------------
--------------  OneShotTimer Timers Module  ----------------
---------------------------------------------------------
do
  local OTE, ODD
  local MTFUNC = function() end

  OneShotTimer = {}

  -- Check to see if new C4: SetTimer functionality exists.  If so, use that instead of 'monkey patching' the OnTimerExpired...
  if (LuaC4Object.SetTimer) then
    local timers = {}

    function OneShotTimer.Callback(timer)
      local tvals = timers[timer]
      if (tvals) then
        timer:Cancel()
        tvals.callback(tvals.name)
        timers[timer] = nil
      end
    end

    function OneShotTimer.ClearAll()
      for timer in pairs(timers) do timer:Cancel() end
      timers = {}
    end

    function OneShotTimer.OnDriverDestroyed()
      OneShotTimer.ClearAll()
      ODD()
    end

    function OneShotTimer.Kill(Name)
      if ((Name or "") == "") then return end
      for timer, tvals in pairs(timers) do
        if (tvals.name == Name) then
          timer:Cancel()
	  timers[timer] = nil
	  return
	end
      end
    end

    INTERVAL_MULT = {MILLISECONDS = 1, SECONDS = 1000, MINUTES = 1000*60, HOURS = 1000*60*60}

    function OneShotTimer.Add(nInterval, strUnits, fCallback, Name)
      OneShotTimer.Kill(Name)
      local time_ms = (tonumber(nInterval) or 1) * (INTERVAL_MULT[strUnits] or 1)
      if (time_ms < 1) then time_ms = 1 end
      local timer = C4:SetTimer(time_ms, OneShotTimer.Callback)
      timers[timer] = {callback = fCallback, name = (Name or "")}
    end

    -----  Overrides:  --------
    ODD , OnDriverDestroyed = OnDriverDestroyed or MTFUNC, OneShotTimer.OnDriverDestroyed
    ---  END Overrides:  ------

  else

    local tCallback = {}
  
    function OneShotTimer.OnTimerExpired(idTimer)
      if (tCallback[idTimer] ~= nil) then
        C4:KillTimer(idTimer)
        tCallback[idTimer].CALLBACK(tostring(tCallback[idTimer].NAME))
        tCallback[idTimer] = nil
      else
        OTE(idTimer)
      end
    end

    function OneShotTimer.ClearAll()
      for k,v in pairs(tCallback) do C4:KillTimer(k) end
      tCallback = {}
    end

    function OneShotTimer.OnDriverDestroyed()
      OneShotTimer.ClearAll()
      ODD()
    end

    function OneShotTimer.Kill(Name)
      if (Name ~= nil) then
        for k,v in pairs(tCallback) do if (v.NAME == Name) then C4:KillTimer(k) tCallback[k] = nil end end
      end
    end

    function OneShotTimer.Add(nInterval, strUnits, fCallback, Name)
      local id = C4:AddTimer(nInterval, strUnits)

      -- Look for name if not nil, if found, remove existing timer callback...
      if (Name ~= nil) then
        for k,v in pairs(tCallback) do if (v.NAME == Name) then C4:KillTimer(k) tCallback[k] = nil end end
      end
      tCallback[id] = {CALLBACK = fCallback, NAME = Name}
    end

    -----  Overrides:  --------
    OTE , OnTimerExpired = OnTimerExpired or MTFUNC, OneShotTimer.OnTimerExpired
    ODD , OnDriverDestroyed = OnDriverDestroyed or MTFUNC, OneShotTimer.OnDriverDestroyed
    ---  END Overrides:  ------

  end
end
---------------------------------------------------------
-----------  END OneShotTimer Timers Module  ---------------
---------------------------------------------------------

--[[FILE_NO_INSERT('ColorDumper.lua')]]
------------------------------------------------------------


-- Enable for debug (and include ColorDumper above)...
--[[
function dump(...) -- luacheck: ignore
  dbg(ColorDumper.dump(...)) -- luacheck: ignore
end
]]


print("Driver Loaded..." .. os.date())


-----------------------------------------------------
------------------------ INIT ------------------------
-----------------------------------------------------
OnPropertyChanged("Debug Mode")
gDev = {}

